import PointPreamble._
import RichNodePreamble._
import GUIPayloadPreamble._
import math._
import ProcessingWrapper._
import collection.mutable.HashMap
import processing.core.PConstants._

/**
 * Created by IntelliJ IDEA.
 * User: Sebastian Stabinger
 * Date: May 16, 2010
 * Time: 6:07:37 PM
 */

class GUIDirectedGraph(val graph: DirectedGraph, val x: Int, val y: Int) {
  val forceDirectedPlacement = new ForceDirectedPlacement(this, x - 220, y)
  var markedPath = List[(Node[GUIPayload], Node[GUIPayload])]()
  var markedNode = Map[Node[GUIPayload], Int]()
  var hiddenProperties = List[(Node[GUIPayload], String)]()
  var markedProperty: (Node[GUIPayload], String) = (null, null)
  var colorBackgroundProperty: Map[(Node[GUIPayload], String), Int] = Map.empty
  val nodeRadius = 15

  val propertyTextSize = 16
  val propertySpacing = 35 // Pixels between Properties of Nodes
  val spacing = 18 // Pixels between Lines of Text
  val propertyPos = Point(x - 200, 20)
  var nodesInProperties = graph.toList
  var selectedProperty: (Node[GUIPayload], String) = (null, "")

  for (i <- 1 to 200) forceDirectedPlacement.iterate()

  /*var forceDirectedPlacementThread = new Thread(new ForceThread())
  forceDirectedPlacementThread.start()*/

  def getNodeAtPosition(x: Int, y: Int, margin: Int): Node[GUIPayload] = {
    for (n <- graph.nodes) {
      if (abs(x - n.pos.x) < margin && abs(y - n.pos.y) < margin) return n
    }
    null
  }

  /**
   *  Returns all Nodes which are in an Area defined by a point an a radius
   */
  def getNodesInArea(x: Int, y: Int, radius: Int) = {
    graph.nodes.filter({n => (n.pos distanceTo (x, y)) <= radius})
  }

  def arrow(n1: Node[GUIPayload], n2: Node[GUIPayload]) {
    ProcessingWrapper.arrow(n1.pos.x, n1.pos.y, n2.pos.x, n2.pos.y, 15)
  }

  def drawNode(n: Node[GUIPayload], textColor: Int, strokeColor: Int, fillColor: Int) {
    p.stroke(strokeColor)
    p.fill(fillColor)
    p.ellipse(n.pos.x, n.pos.y, 2 * nodeRadius, 2 * nodeRadius)
    p.fill(textColor)
    p.text(n.name, n.pos.x, n.pos.y - 3)
  }

  def drawEdge(n1: Node[GUIPayload], n2: Node[GUIPayload], color: Int) {
    p.stroke(color)
    if (n1 != n2)
      arrow(n1, n2)
    else
      drawSelfeEdge(n1)
  }

  def draw(n: Node[GUIPayload]) {
    drawNode(n,
      if (n.isNF) NODE_NF_TEXT_COLOR else NODE_TEXT_COLOR,
      if (markedPath exists {x => x._1 == n || x._2 == n}) MARKED_PATH_COLOR else NODE_BORDER_COLOR,
      if (markedNode contains n) NODE_MARKED_FILL_COLOR else if (markedProperty._1 == n) RED else NODE_FILL_COLOR)

    for (i <- n)
      drawEdge(n, i, if (markedPath contains (n, i)) RED else LINECOLOR)
  }

  def gotIntersections: Boolean = {
    for (n1 <- graph.connectedPairs; n2 <- graph.connectedPairs if (n1 != n2))
      if (Line(n1._1, n1._2) intersects Line(n2._1, n2._2)) return true
    false
  }

  def markHiddenProperty() {
    markedProperty = hiddenProperties(GetRandom(hiddenProperties.length))
  }

  def checkPropertyString(prop: String) {
    require(prop == "SN" || prop == "WN" || prop == "CR" || prop == "WCR" || prop == "UN")
  }

  def hideNodeProperty(n: Node[GUIPayload], prop: String) {
    checkPropertyString(prop)
    hiddenProperties = (n, prop) :: hiddenProperties
  }

  def hideAllNodeProperties(n: Node[GUIPayload]) = List("SN", "WN", "CR", "WCR", "UN") foreach {hideNodeProperty(n, _)}

  def hideAllProperties() = graph foreach {hideAllNodeProperties(_)}

  def showNodeProperty(n: Node[GUIPayload], prop: String) {
    checkPropertyString(prop)
    hiddenProperties = hiddenProperties filterNot {_ == (n, prop)}
  }

  def showAllNodeProperties(n: Node[GUIPayload]) = List("SN", "WN", "CR", "WCR", "UN") foreach {showNodeProperty(n, _)}

  def showAllProperties() {
    hiddenProperties = Nil
    markedProperty = (null, null)
  }

  /**
   * Draws the properties of one node to the screen beginning at pos
   */
  def drawNodeProperties(n: Node[GUIPayload], pos: Point) {

    val propertyList = List("SN", "WN", "CR", "WCR", "UN")

    p.fill(TEXTCOLOR)
    p.text(n.name, pos.x, pos.y)

    def drawOneProp(prop: String, offset: Int) {

      def currentPropertyHolds = {
        prop match {
          case "SN" => n.isSN
          case "WN" => n.isWN
          case "CR" => n.isCR
          case "WCR" => n.isWCR
          case "UN" => n.isUN
          case _ => error("Property not known: " + prop)
        }
      }

      // Paints the background of this property a certain color
      def dyeBackground(color: Int) = {
        p.fill(color)
        p.stroke(color)
        p.rect(pos.x - propertySpacing / 2 + (offset * propertySpacing), (pos.y - spacing / 2) + 2, propertySpacing, spacing)
      }
      // Color Background if necessary
      colorBackgroundProperty.get((n, prop)) match { // In Game for right/wrong answers
        case Some(color) => dyeBackground(color)
        case None => ()
      }
      if (selectedProperty == (n, prop)) dyeBackground(MARKED_PROPERTY_COLOR) // In Edit if a property is selected

      p.image(
        if (markedProperty == (n, prop))
          astImg
        else if (hiddenProperties contains (n, prop))
          questionImg
        else if (currentPropertyHolds)
          tickImg
        else
          crossImg,
        pos.x + (offset * propertySpacing), pos.y + 2)
    }
    for ((p, i) <- propertyList.zipWithIndex) drawOneProp(p, i + 1)
  }

  /**
   * Draws the properties of all nodes
   */
  def drawNodeProperties(pos: Point) {
    p.textSize(propertyTextSize)
    p.imageMode(CENTER)
    var ypos = pos.y + spacing
    nodesInProperties = (graph.toList).sortWith(_.name < _.name)

    // Draw Legend
    p.fill(TEXTCOLOR)
    p.text("SN", pos.x + propertySpacing, pos.y)
    p.text("WN", pos.x + 2 * propertySpacing, pos.y)
    p.text("CR", pos.x + 3 * propertySpacing, pos.y)
    p.text("WCR", pos.x + 4 * propertySpacing, pos.y)
    p.text("UN", pos.x + 5 * propertySpacing, pos.y)

    // Draw Properties
    for (node <- nodesInProperties) {
      drawNodeProperties(node, (pos.x, ypos))
      ypos = ypos + spacing
    }
    p.textSize(standardTextSize)
  }

  def propertyBoundingBox = {
    require(propertyPos != null, "Tried to get bounding box of properties with pos == null")
    val p1: Point = (propertyPos.x - (propertyTextSize / 2), propertyPos.y + (propertyTextSize / 2)) // Upper Left Point
    val p2: Point = (p1.x + 6 * propertySpacing /*+ (propertySpacing / 2)*/ , p1.y + graph.size * spacing) // Lower Right Point
    (p1, p2)
  }

  /**
   * Returns a box surrounding all nodes
   */
  def boundingBox: Box = {
    if (graph.nonEmpty) {
      var left = Integer.MAX_VALUE
      var right = Integer.MIN_VALUE
      var top = Integer.MAX_VALUE
      var bottom = Integer.MIN_VALUE
      for (i <- graph) {
        left = left min i.pos.x
        right = right max i.pos.x
        top = top min i.pos.y
        bottom = bottom max i.pos.y
      }
      Box(Point(left, top), Point(right, bottom))
    }
    else
      Box(Point(0, 0), Point(0, 0))
  }

  /**
   * Checks whether pos lies inside the propertybox or not
   */
  def isInPropertyBox(pos: Point) = {
    val (p1, p2) = propertyBoundingBox
    pos between (p1, p2)
  }

  /**
   * Returns the selected node in the propertybox
   * @return The selected node
   */
  def getNodeOnPropertyPosition(pos: Point) = {
    if (!isInPropertyBox(pos))
      null
    else {
      val y = pos.y - propertyPos.y
      val tmp = (y - (spacing / 2) - 4) / spacing
      nodesInProperties(tmp min (nodesInProperties.size - 1))
    }
  }

  /**
   * Returns the property selected in the propertybox
   */
  def getSelectedProperty(pos: Point) = {
    if (!isInPropertyBox(pos))
      "none"
    else {
      val props = List("Name", "SN", "WN", "CR", "WCR", "UN")
      val x = pos.x - propertyPos.x
      val tmp = x / propertySpacing
      if (tmp > 5 || tmp < 0) "none" else props(tmp)
    }
  }

  def drawExampleForProperty(node: Node[GUIPayload], prop: String) {
    if (node != null) {
      markedNode = Map(node -> BLUE)
      markedPath = prop match {
        case "SN" => if (node.isSN) Nil else node.cycles(0)
        case "WN" => {
          if (node.isWN) {
            if (node.isNF) List((node, node)) else {
              // Creates a sorted (by length of path) List of Paths to all NormalForms
              val pathsToNF = node.getNF.map(node.shortestPathTo(_)).sortWith(_.length < _.length)
              pathsToNF(0) // Coose the shortest Path
            }
          } else node.getTransitivePath
        }
        case "WCR" =>
          if (node.isWCR) Nil else {
            val counter = node.getWcrCounterexamples(0)
            node.shortestPathTo(counter._1) union node.shortestPathTo(counter._2)
          }
        case "CR" =>
          if (node.isCR) Nil else {
            val counter = node.getCrCounterexample(0)
            node.shortestPathTo(List(counter._1, counter._2))
          }
        case "UN" =>
          if (node.isUN) node.getTransitivePath else node.shortestPathTo(node.getNF)
        case _ => Nil
      }
    }
  }

  /**
   * Marks pieces of the graph to illustrate a property which is clicked on
   */
  def drawExampleForSelectedProperty(pos: Point) {
    if (isInPropertyBox(pos)) {
      val node = getNodeOnPropertyPosition(pos)
      val prop = getSelectedProperty(pos)
      drawExampleForProperty(node, prop)
      selectedProperty = (node, prop)
    }
  }

  /**
   * Draws the edge from a node to the same node
   */
  def drawSelfeEdge(node: Node[GUIPayload]) = {
    p.noFill()
    //p.stroke(LINECOLOR)
    val center = node.pos
    val xOffset = 10
    val yOffset = 15
    val width = 25
    val height = 50
    p.bezier(center.x - xOffset, center.y - yOffset,
      center.x - width, center.y - height, center.x + width, center.y - height,
      center.x + xOffset, center.y - yOffset)
    // Get 2 Points at the end of the beziercurve to draw the arrowhead
    val point1x = p.bezierPoint(center.x - xOffset, center.x - width, center.x + width, center.x + xOffset, 0.9F)
    val point1y = p.bezierPoint(center.y - yOffset, center.y - height, center.y - height, center.y - yOffset, 0.9F)
    val point2x = p.bezierPoint(center.x - xOffset, center.x - width, center.x + width, center.x + xOffset, 1.0F)
    val point2y = p.bezierPoint(center.y - yOffset, center.y - height, center.y - height, center.y - yOffset, 1.0F)
    ProcessingWrapper.arrow(point1x.toInt, point1y.toInt, point2x.toInt, point2y.toInt, 0)
  }

  /**
   * Returns a list of pairs of pairs of nodes which intersect
   */
  def intersections = {
    // Create list of pairs of pairs without doubles
    var combinations: List[((Node[GUIPayload], Node[GUIPayload]), (Node[GUIPayload], Node[GUIPayload]))] = Nil
    for (i <- graph.connectedPairs; j <- graph.connectedPairs if (!(combinations contains (j, i)) && i != j))
      combinations = (i, j) :: combinations
    // Filter the list for intersections
    combinations.filter(x => Line(x._1) intersects Line(x._2))
  }

  def draw() {
    drawExampleForProperty(selectedProperty._1, selectedProperty._2)
    graph.foreach(draw(_))
  }

  override def toString() = {
    def nodeIdentifier(n: Node[GUIPayload]) = n.name + "(" + n.pos.x + "," + n.pos.y + ")" // Returns name(x,y) ex. x(234,567)
    val nodelist = graph.nodes.filter(n => n.connections.nonEmpty || !graph.isConnected(n))
    nodelist.map(n => nodeIdentifier(n) + "->" + n.connections.map(_.name).mkString(",")).mkString("\n")
  }
}