import ProcessingWrapper._
import PointPreamble._

/**
 * Implements the gamelogic of the tool
 */
class GameDialog(val applet: AbstractRewriteTool) {
  val guiGraph = applet.guiGraph
  val messageSystem = applet.messageSystem

  var yesButton: PImageButton = null
  var noButton: PImageButton = null
  refresh()

  var active = true

  val yesXOffset = 50
  val yesYOffset = 50
  val noXOffset = 160
  val noYOffset = 50

  var mistakes = 0

  var stopWatchStartTime: Long = 0
  var finishTime: Long = 0

  def pos = {
    val (p1, p2) = guiGraph.propertyBoundingBox
    (p1.x, p2.y + 30)
  }

  def textPos = pos + (yesXOffset + (noXOffset - yesXOffset) / 2, 10)

  def somethingMarked = (guiGraph.markedProperty != (null, null))

  def systemSeconds = System.nanoTime / 1000000000L

  def startStopWatch() = if (stopWatchStartTime == 0) stopWatchStartTime = systemSeconds

  def stopStopWatch() = if (finishTime == 0) finishTime = stopWatchTime

  def stopWatchTime = if (stopWatchStartTime != 0) systemSeconds - stopWatchStartTime else 0

  def stopWatchString = {
    var time = if (finishTime == 0) stopWatchTime else finishTime
    val sec = time % 60
    time / 60 + ":" + (if (sec < 10) "0" + sec else sec.toString)
  }

  def correctAnswer[A](n: Node[A], prop: String) = {
    prop match {
      case "SN" => n.isSN
      case "WN" => n.isWN
      case "CR" => n.isCR
      case "WCR" => n.isWCR
      case "UN" => n.isUN
      case _ => error("Tried to match Property: " + prop + " @ node: " + n); false
    }
  }

  val mSize = 15
  val mDuration = 1000

  def yesMarkWrongAnswer(n: Node[GUIPayload], prop: String) {
    prop match {
      case "SN" =>
        messageSystem.addMessage("Wrong: There exists a cycle (see marked path in graph)", mDuration, RED, mSize)
        messageSystem.addMessage("\u21d2 Not Strongly Normalizing", mDuration, RED, mSize)
        guiGraph.drawExampleForProperty(n, prop)
      case "WN" =>
        messageSystem.addMessage("Wrong: There is no normal form reachable from this node", mDuration, RED, mSize)
        messageSystem.addMessage("(All reachable nodes are marked in the graph)", mDuration, RED, mSize)
        messageSystem.addMessage("\u21d2 Not Weakly Normalizing", mDuration, RED, mSize)
        guiGraph.drawExampleForProperty(n, prop)
      case "CR" =>
        messageSystem.addMessage("Wrong: There are nodes that can't be joined (see marked nodes in graph)", mDuration, RED, mSize)
        messageSystem.addMessage("\u21d2 Not Church Rosser", mDuration, RED, mSize)
        guiGraph.drawExampleForProperty(n, prop)
      case "WCR" =>
        messageSystem.addMessage("Wrong: There are nodes, that can't be joined (see marked nodes in graph)", mDuration, RED, mSize)
        messageSystem.addMessage("\u21d2 Not Weak Church Rosser", mDuration, RED, mSize)
        guiGraph.drawExampleForProperty(n, prop)
      case "UN" =>
        messageSystem.addMessage("Wrong: There is more than one normal form reachable (see marked nodes in graph)", mDuration, RED, mSize)
        messageSystem.addMessage("\u21d2 No Unique Normal Forms", mDuration, RED, mSize)
        guiGraph.drawExampleForProperty(n, prop)
    }
  }

  def noMarkWrongAnswer(n: Node[GUIPayload], prop: String) {
    prop match {
      case "SN" =>
        messageSystem.addMessage("Wrong: There exists no cycle that is reachable from this node", mDuration, RED, mSize)
        messageSystem.addMessage("\u21d2 Strongly Normalizing", mDuration, RED, mSize)
        guiGraph.drawExampleForProperty(n, prop)
      case "WN" =>
        messageSystem.addMessage("Wrong: There is a normal form reachable from this node", mDuration, RED, mSize)
        messageSystem.addMessage("(A reachable normal form is marked in the graph)", mDuration, RED, mSize)
        messageSystem.addMessage("\u21d2 Weakly Normalizing", mDuration, RED, mSize)
        guiGraph.drawExampleForProperty(n, prop)
      case "CR" =>
        messageSystem.addMessage("Wrong: All nodes reachable from this node are joinable", mDuration, RED, mSize)
        messageSystem.addMessage("\u21d2 Church Rosser", mDuration, RED, mSize)
        guiGraph.drawExampleForProperty(n, prop)
      case "WCR" =>
        messageSystem.addMessage("Wrong: All direct descendants from this node are joinable", mDuration, RED, mSize)
        messageSystem.addMessage("\u21d2 Weak Church Rosser", mDuration, RED, mSize)
        guiGraph.drawExampleForProperty(n, prop)
      case "UN" =>
        messageSystem.addMessage("Wrong: No or one normal form is reachable from this node", mDuration, RED, mSize)
        messageSystem.addMessage("(All reachable nodes are marked in the graph)", mDuration, RED, mSize)
        messageSystem.addMessage("\u21d2 Unique Normal Forms", mDuration, RED, mSize)
        guiGraph.drawExampleForProperty(n, prop)
    }
  }

  def yesWrongAnswer(n: Node[GUIPayload], prop: String) {
    guiGraph.colorBackgroundProperty += (n, prop) -> WRONG_MARK_COLOR
    messageSystem.clear()
    yesMarkWrongAnswer(n, prop)
    mistakes += 1
  }

  def noWrongAnswer(n: Node[GUIPayload], prop: String) {
    guiGraph.colorBackgroundProperty += (n, prop) -> WRONG_MARK_COLOR
    messageSystem.clear()
    noMarkWrongAnswer(n, prop)
    mistakes += 1
  }

  def rightAnswer(n: Node[GUIPayload], prop: String) {
    guiGraph.colorBackgroundProperty += (n, prop) -> RIGHT_MARK_COLOR
    messageSystem.clear()
    messageSystem.addMessage("Right Answer!", 5, GREEN)
    guiGraph.markedPath = Nil
    guiGraph.markedNode = Map.empty
  }

  def refresh() {
    val (n, prop) = guiGraph.markedProperty
    def commonPart() {
      guiGraph.showNodeProperty(n, prop)
      guiGraph.markHiddenProperty()
      if (guiGraph.hiddenProperties.size > 1)
        refresh()
      else {
        guiGraph.hiddenProperties = Nil
        guiGraph.markedProperty = (null, null)
        active = false
      }
    }

    yesButton = PImageButton(acceptImg, pos + (yesXOffset, yesYOffset)) {
      if (correctAnswer(n, prop))
        rightAnswer(n, prop)
      else
        yesWrongAnswer(n, prop)
      commonPart()
    }

    noButton = PImageButton(declineImg, pos + (noXOffset, noYOffset)) {
      if (!correctAnswer(n, prop))
        rightAnswer(n, prop)
      else
        noWrongAnswer(n, prop)
      commonPart()
    }
  }

  def draw() {
    if (active) {
      def printText() {
        p.textSize(30)
        p.text(guiGraph.markedProperty._2, textPos.x, textPos.y)
        p.textSize(standardTextSize)
      }
      def printMistakes() {
        val x = guiGraph.propertyBoundingBox._1.x / 2
        p.textSize(30)
        p.fill(MISTAKE_COUNTER_COLOR)
        p.text("Mistakes: " + mistakes + " out of " + guiGraph.graph.size * 5, x, 40)
        p.fill(TIME_COUNTER_COLOR)
        p.text(stopWatchString, x, 40 + 30)
        p.fill(WHITE)
        p.textSize(standardTextSize)
      }

      printMistakes()
      /*if (somethingMarked) {
        printText()
        yesButton.draw()
        noButton.draw()
      }*/
    }
  }

  def check(mouse: Point) {
    if (active) {
      yesButton.check(mouse)
      noButton.check(mouse)
    }
  }
}