import collection.mutable.HashMap
import scala.util.Random
import PointPreamble._
import GUIPayloadPreamble._

/**
 * Created by IntelliJ IDEA.
 * User: Sebastian Stabinger
 * Date: Apr 5, 2010
 * Time: 12:39:01 AM
 */

// Am besten Random erzeugte Graphen automatisch untersuchen
object GetRandom {
  val rnd = new Random

  def apply() = {
    rnd.nextInt(700)
  }

  def apply(until: Int) = rnd.nextInt(until)
}

/**
 * Implements some static methods to work with graphs
 */
object DirectedGraph {
  def getGraphFromFile(file: String): Option[DirectedGraph] = {
    try {
      val g = new MemDirectedGraph
      for (i <- ARTFileParser(file)) {
        g += i
      }
      Some(g)
    } catch {
      case ex: RuntimeException => None
    }
  }

  def getGraphFromString(s: String): Option[DirectedGraph] = {
    try {
      ARTStringParser(s) match {
        case Some(x) =>
          val g = new MemDirectedGraph
          for (i <- x) g += i
          Some(g)
        case None => None
      }
    } catch {
      case ex: RuntimeException => None
    }
  }


  def getRandomGraph(nodes: Int, vertices: Int): DirectedGraph = {
    val res = new MemDirectedGraph
    val rnd = new Random
    for (i <- 1 to (nodes * vertices)+2)
      res += (rnd.nextInt(nodes).toString, List(rnd.nextInt(nodes).toString))
    res.connectGraph()
    res.renumberNodes()
    res
  }

  def getNonMemRandomGraph(nodes: Int, vertices: Int): DirectedGraph = {
    val res = new DirectedGraph
    val rnd = new Random
    for (i <- 1 to (nodes * vertices))
      res += (rnd.nextInt(nodes).toString, List(rnd.nextInt(nodes).toString))
    res.connectGraph()
    res.renumberNodes()
    res
  }

  def getCompleteGraph(nodes: Int) = {
    var res = new MemDirectedGraph
    for (i <- 1 to nodes) res += (i.toString, Nil) // Create Nodes
    for (n <- res; m <- res if (n != m)) n += m // Create connections from all nodes to all nodes
    res
  }
}

class DirectedGraph extends Iterable[Node[GUIPayload]] {
  var nodes = new NodeList

  // Add a new Node to the Graph Format: Tuple(From,List(To,To,...))
  def +=(node: (String, List[String])) = {
    val from = nodes(node._1)
    from.pos = Point(GetRandom(), GetRandom())
    for (to <- node._2) from += nodes(to)
  }

  // Delete a Node from the Graph
  def -=(node: Node[GUIPayload]) = {
    for (i <- nodes) i -= node
    nodes.hash -= node.name
  }

  def renumberNodes() {
    val nodeList = nodes.toList
    var nodeMap = Map[String, String]()
    def createNodeMap() {
      var count = 0
      for (i <- nodeList) {
        nodeMap = nodeMap + (i.name -> count.toString)
        count += 1
      }
    }
    createNodeMap()
    nodes = new NodeList
    for (i <- nodeList) {
      var conString = (for (c <- i.connections) yield nodeMap(c.name)).toList
      this += (nodeMap(i.name) -> conString)
    }
  }

  // Add a new Edge. Implemented mainly because of memoizing
  def addEdge(from: Node[GUIPayload], to: Node[GUIPayload]) = from += to
  // Removes an Edge
  def removeEdge(from: Node[GUIPayload], to: Node[GUIPayload]) = from -= to

  override def size = nodes.size

  def apply(name: String) = if (nodes.hash.contains(name)) nodes(name) else null

  def getNF = (for (i <- nodes if (i.isNF)) yield i).toList

  // returns true if there is a connection from/to n
  def isConnected(n: Node[GUIPayload]): Boolean = {
    if (n.connections.nonEmpty) true
    else {
      nodes.flatMap(_.connections).exists(_ == n) // Look if there is a connection to n
    }
  }

  /**
   * Returns a mapping of nodes to their transitive closure
   */
  def transitiveClosure = {
    import collection.mutable.Map
    // Create adjacencyMatrix
    val A: Map[Node[GUIPayload], Map[Node[GUIPayload], Boolean]] = Map.empty
    for (i <- nodes) A += (i -> Map.empty)
    for (i <- nodes; j <- nodes) {
      A(i) += (j -> (i.connections contains j))
    }
    // Warshall's Algorithm
    val T = A.clone
    for (j <- nodes; i <- nodes if (T(i)(j) == true)) {
      for (k <- nodes) T(i)(k) = T(i)(k) || T(j)(k)
    }
    // Create Datastructure
    import collection.mutable.Set
    val res: Map[Node[GUIPayload], Set[Node[GUIPayload]]] = Map.empty
    for (i <- nodes) res += (i -> Set.empty)
    for (j <- nodes; i <- nodes if (T(j)(i))) {
      res(j) += i
    }
    res
  }

  /**
   * Returns a List of Lists containing Nodes which are part of strongly connected components
   * Uses Gabrow's algorithm (http://en.wikipedia.org/wiki/Gabow's_algorithm)
   */
  def stronglyConnectedComponents(): List[List[Node[GUIPayload]]] = {
    import collection.mutable.Stack
    import collection.mutable.Map

    var components: List[List[Node[GUIPayload]]] = Nil

    type NodeList = List[Node[GUIPayload]]

    var C = 0
    val S = new Stack[Node[GUIPayload]]
    val P = new Stack[Node[GUIPayload]]
    var preOrder: Map[Node[GUIPayload], Int] = Map.empty

    // Returns true if the node already is in a component
    def inComponent(n: Node[GUIPayload]) = components.exists(_.exists(_ == n))

    def gabow(v: Node[GUIPayload]) {
      preOrder(v) = C
      C += 1
      S.push(v)
      P.push(v)
      for (w <- v.connections) {
        preOrder.get(w) match {
          case None =>
            gabow(w)
          case Some(prew) =>
            if (!inComponent(w)) {
              while ( /*P.nonEmpty && */ prew < preOrder.getOrElse(P.top, Integer.MAX_VALUE))
                P.pop()
            }
        }
      }
      if ( /*P.nonEmpty && */ v == P.top) {
        var component: NodeList = Nil
        var popped: Node[GUIPayload] = null
        while (popped != v) {
          popped = S.pop()
          component = popped :: component
        }
        P.pop()
        components = component :: components
      }
    }
    // Main Loop
    for (n <- nodes) {
      if (preOrder.contains(n)) () else gabow(n)
    }
    components
  }

  //TODO:Only for Testing
  def newCR() = {
    import collection.mutable.{Map, Set}
    val mark: Map[Node[GUIPayload], Int] = Map.empty
    var markCount = 0
    val set: Set[Node[GUIPayload]] = Set.empty
    for (i <- nodes if (i.isNF || i.connections.contains(i))) {
      mark += (i -> markCount)
      set += i
      markCount += 1
    }
    def iteration() {
      for (i <- set) {
        for (j <- nodes.filter(n => n.connections.contains(i))) {
          if (mark.contains(j)) {
            if (mark(i) != mark.getOrElse(i, -1)) {
              mark(j) = -1
            }
          }
          else {
            mark += (j -> mark(i))
            set += j
          }
        }
      }
    }
    for (i <- 1 to 100) iteration()
    mark
  }

  /**
   * Returns a set of all nodes which are part of a cycle
   */
  def cycleNodes: Set[Node[GUIPayload]] = {
    import collection.mutable.HashSet

    val selfReferenzing = nodes.filter(n => n.connections.contains(n)).map(_.asInstanceOf[Node[GUIPayload]]).toList
    val cycles = if (selfReferenzing == Nil)
      stronglyConnectedComponents.filter(_.length > 1)
    else
      selfReferenzing :: stronglyConnectedComponents.filter(_.length > 1)

    var res = Set[Node[GUIPayload]]()
    cycles.foreach(_.foreach(res += _))
    res
  }

  // Experimental algorithm to decide SN
  def expSN() = {
    import collection.mutable.Map
    val sn: Map[Node[GUIPayload], Boolean] = Map.empty

    def visit(n: Node[GUIPayload], visited: Set[Node[GUIPayload]]) {
      if (!(sn contains n)) {
        if (visited contains n) sn += (n -> false)
        else {
          for (i <- n.connections) visit(i, visited + n)
          sn += (n -> (n.connections.forall(sn(_) == true)))
        }
      }
    }

    for (i <- nodes if (!sn.contains(i))) visit(i, Set.empty)
    sn
  }

  /**
   * Returns all nodes which are not strongly normalizing
   */
  def notSnNodes = {
    val cycles = cycleNodes
    val closure = transitiveClosure
    nodes.toList.filter(x => cycles.contains(x) || closure(x).exists(cycles.contains(_)))
  }

  /**
   * Returns a list of lists of nodes which are part of unconnected graphs.
   * If all nodes of the graph are connected only one list will be returned
   */
  def disconnectedSubgraphs: List[List[Node[GUIPayload]]] = {
    var subgraphs = List[List[Node[GUIPayload]]]()
    val closure = transitiveClosure
    for (i <- nodes) subgraphs = (i :: closure(i).toList) :: subgraphs
    var change = true
    while (change) {
      change = false
      for (i <- subgraphs; j <- subgraphs; val int = i intersect j) {
        if (i != j && int.length != 0) {
          subgraphs = (i union j) :: subgraphs.filterNot(x => x == i || x == j)
          change = true
        }
      }
      subgraphs = (for (i <- subgraphs) yield i.distinct).distinct
    }
    subgraphs
  }

  /**
   * All unconnected parts of the graph will be connected
   */
  def connectGraph() {
    val dis = disconnectedSubgraphs
    if (dis.length > 1) for (i <- dis if (i != dis(0))) {
      addEdge(i(0), dis(0)(0))
    }
  }

  /**
   * Returns a list of pairs of all nodes
   */
  def pairs = {
    var list: List[(Node[GUIPayload], Node[GUIPayload])] = Nil
    for (i <- nodes; j <- nodes if (!((list contains (j, i)) || i == j))) list = (i, j) :: list
    list
  }

  /**
   * Returns a list of pairs of nodes which are connected without nodes connected to themself.
   * Two nodes connected by two edges in different directions will only produce one pair
   */
  def connectedPairs = {
    var connectedPairsTmp: List[(Node[GUIPayload], Node[GUIPayload])] = Nil
    for (i <- pairs if ((i._1.neighbors contains i._2) || (i._2.neighbors contains i._1)))
      if (!(connectedPairsTmp contains (i._2, i._1))) connectedPairsTmp = i :: connectedPairsTmp
    connectedPairsTmp
  }

  def iterator = nodes.iterator

  override def toString() = nodes.filter(n => n.connections.nonEmpty || !isConnected(n)).mkString("\n")
}

/**
 * Provides functionality to manage the list of nodes
 */
class NodeList extends Iterable[MemNode[GUIPayload]] {
  val hash = new HashMap[String, MemNode[GUIPayload]]()

  override def size = hash.size

  def apply(key: String): MemNode[GUIPayload] = hash.getOrElseUpdate(key, new MemNode(key, new GUIPayload))

  def iterator = hash.iterator.map(_._2)
}

/**
 *  Implements all Memoization for Graph
 */
class MemDirectedGraph extends DirectedGraph {
  def clearMem() {
    memPairs = null
    memIsConnected = Memo.getFunctionHashMap(isConnected)
    memGetNF = null
    memConnectedPairs = null
    for (i <- nodes) i.clearMem()
  }

  var memPairs: List[(Node[GUIPayload], Node[GUIPayload])] = null

  override def pairs = {
    if (memPairs == null) memPairs = super.pairs
    memPairs
  }

  var memIsConnected = Memo.getFunctionHashMap(isConnected)

  override def isConnected(n: Node[GUIPayload]) = memIsConnected.getOrElseUpdate(n, super.isConnected(n))

  var memGetNF: List[MemNode[GUIPayload]] = null

  override def getNF = {
    if (memGetNF == null) memGetNF = super.getNF
    memGetNF
  }

  var memConnectedPairs: List[(Node[GUIPayload], Node[GUIPayload])] = null

  override def connectedPairs = {
    if (memConnectedPairs == null) memConnectedPairs = super.connectedPairs
    memConnectedPairs
  }


  override def -=(node: Node[GUIPayload]) = {
    clearMem()
    super.-=(node)
  }

  override def +=(node: (String, List[String])) = {
    clearMem()
    super.+=(node)
  }

  override def removeEdge(from: Node[GUIPayload], to: Node[GUIPayload]) {
    clearMem()
    super.removeEdge(from, to)
  }

  override def addEdge(from: Node[GUIPayload], to: Node[GUIPayload]) {
    clearMem()
    super.addEdge(from, to)
  }
}