import processing.core.PConstants._
import ProcessingWrapper._
import PointPreamble._
import GUIPayloadPreamble._
import RichNodePreamble._
import collection.mutable.HashMap
import Memo._

class Node[T](val name: String, var payload: T) extends Iterable[Node[T]] {
  var connections: List[Node[T]] = Nil

  /*var closureMem: List[Node[T]] = Nil
  var crCounterExamplesMem: List[(Node[T], Node[T])] = Nil
  var joiningNodesMem = HashMap[Node[T], List[Node[T]]]()*/

  // Adds a connection to another node
  @deprecated("Use addEdge in DirectedGraph instead")
  def +=(node: Node[T]) = {
    if (!(connections contains node)) connections = node :: connections
  }
  // Removes a connection to another node
  @deprecated("Use removeEdge in DirectedGraph instead")
  def -=(node: Node[T]) = {
    connections = connections filterNot {_ == node}
  }
  // Returns the same node with different name
  def rename(name: String) = {
    val ret = new Node[T](name, payload)
    ret.connections = this.connections
    ret
  }
  // Returns the list of nodes reachable in one step
  def neighbors = connections
  // Returns list of all nodes reachable from this node
  def closure: List[Node[T]] = reaches(Nil)
  // Recursive method used in reaches
  def reaches(visited: List[Node[T]]): List[Node[T]] = {
    var tmpvis = visited
    for (i <- connections if (!tmpvis.contains(i))) {
      tmpvis = tmpvis union i.reaches(List(i) union tmpvis)
    }
    tmpvis
  }
  // Returns true if this can be rewritten to node
  def rewritesTo(node: Node[T]) = closure contains node
  // Returns ture if this is a reduct of node
  def reductOf(node: Node[T]) = closure contains this
  // Returns all nodes over witch this and node are joinable
  def getJoiningNodes(node: Node[T]) = (this :: closure) intersect (node :: node.closure)


  // Returns true when this and node are joinable
  def joines(node: Node[T]) = (this getJoiningNodes node).size != 0
  // Get all normal forms of this
  def getNF = if (isNF) List(this) else closure.filter(_.isNF).distinct
  // Returns true if this is in normal form
  def isNF = (connections.size == 0)
  // Return true if this is Stronly Normalizing
  def isSN: Boolean = isSN(Nil);
  // Recursive method used in isSN
  private def isSN(visited: List[Node[T]]): Boolean = {
    for (i <- connections) {
      if (visited contains i) return false
      if (i.isSN(List(i) union visited) == false) return false
    }
    true
  }
  // Return true if this is Weakly Normalizing
  def isWN = getNF.size != 0

  // Returns a list of counterexamples for Church Rosser
  def getCrCounterexample: List[(Node[T], Node[T])] = {
    for (i <- closure; j <- closure if (!(i joines j))) return List((i, j))
    Nil
  }

  // Return true if this is Church Rosser
  def isCR = getCrCounterexample == Nil

  // Returns a list of counterexamples for Weak Church Rosser
  def getWcrCounterexamples = {
    for (i <- neighbors; j <- neighbors if (!(i joines j))) yield (i, j)
  }

  // Returns true if this node is WCR
  def isWCR = getWcrCounterexamples == Nil

  // Return true if this is Unique Normalizing
  def isUN = getNF.size <= 1

  // Returns shortest Path from this to Node n
  def shortestPathTo(n: Node[T]): List[(Node[T], Node[T])] = {
    // TODO: Algorithmus verbessern
    var visited = List(this)
    var l1: List[List[Node[T]]] = List(List(this))

    do {
      l1 = for (i <- l1; j = i.head; k <- j if (!(visited contains k))) yield {
        visited = k :: visited
        k :: i
      }
    } while (!l1.isEmpty && l1.count({x => x.head == n}) == 0)
    if (l1 isEmpty)
      List[(Node[T], Node[T])]()
    else {
      val tmp = l1.filter({x => x.head == n}).head.reverse
      tmp zip tmp.tail
    }
  }

  // Returns a concatenation of all edges of shortest Paths to a List of Nodes
  def shortestPathTo(n: List[Node[T]]): List[(Node[T], Node[T])] = {
    n.flatMap(shortestPathTo(_)).distinct
  }

  // Returns all edges needed to reach all reachable nodes
  def getTransitivePath = shortestPathTo(closure)

  // Returns all cycles reachable from this node
  def cycles = {
    //TODO: Schöneren Algorithmus schreiben
    var result = List[List[Node[T]]]()
    var tmp1 = List(List(this))
    var tmp2 = List[List[Node[T]]]()
    do {
      for (i <- tmp1; j <- i.head) {
        if (i contains j)
          result = (j :: i) :: result
        else
          tmp2 = (j :: i) :: tmp2
      }
      tmp1 = tmp2
      tmp2 = List[List[Node[T]]]()
    } while (!tmp1.isEmpty)
    if (result.isEmpty)
      List[List[(Node[T], Node[T])]]()
    else {
      result.map {
        x =>
          val tmp = x.reverse
          tmp zip tmp.tail
      }
    }
  }

  def iterator = connections.iterator

  override def toString() = {
    if (connections == Nil)
      name
    else
      name + "->" + (connections map {_.name}).mkString(",")
  }
}

object MemNode {
  def apply[T](n: Node[T]) = {
    val mNode = new MemNode(n.name, n.payload)
    mNode.connections = n.connections
    mNode
  }
}

class MemNode[T](name: String, payload: T) extends Node[T](name, payload) {
  def clearMem() {
    memShortestPathTo = getFunctionHashMap(shortestPathTo)
    memRewritesTo = getFunctionHashMap(rewritesTo)
    memReductOf = getFunctionHashMap(reductOf)
    memJoins = getFunctionHashMap(joines)
    memIsWN = None
    memIsWCR = None
    memIsUN = None
    memIsSN = None
    memIsNF = None
    memIsCR = None
    memGetWcrCounterexamples = null
    memGetTransitivePath = null
    memGetNF = null
    memGetJoiningNodes = getFunctionHashMap(getJoiningNodes)
    memGetCrCounterexamples = null
    memCycles = null
    memClosure = null
  }

  var memShortestPathTo = getFunctionHashMap(shortestPathTo)

  override def shortestPathTo(n: Node[T]) = memShortestPathTo.getOrElseUpdate(n, super.shortestPathTo(n))

  var memRewritesTo = getFunctionHashMap(rewritesTo)

  override def rewritesTo(node: Node[T]) = memRewritesTo.getOrElseUpdate(node, super.rewritesTo(node))

  var memReductOf = getFunctionHashMap(reductOf)

  override def reductOf(node: Node[T]) = memReductOf.getOrElseUpdate(node, super.reductOf(node))

  var memJoins = getFunctionHashMap(joines)

  override def joines(node: Node[T]) = memJoins.getOrElseUpdate(node, super.joines(node))

  var memIsWN: Option[Boolean] = None

  override def isWN = {
    if (memIsWN == None) {
      val res = super.isWN
      if(res == false) memIsSN = Some(false)
      memIsWN = Some(res)
    }
    memIsWN.get
  }

var memIsWCR: Option[Boolean] = None

  override def isWCR = {
    if (memIsWCR == None) {
      val res = super.isWCR
      if(res == false) memIsCR = Some(false)
      memIsWCR = Some(res)
    }
    memIsWCR.get
  }

  var memIsUN: Option[Boolean] = None

  override def isUN = {
    if (memIsUN == None) {
      val res = super.isUN
      if(res == false) memIsCR = Some(false)
      memIsUN = Some(res)
    }
    memIsUN.get
  }

  var memIsSN: Option[Boolean] = None

  override def isSN = {
    if (memIsSN == None) {
      val res = super.isSN
      if(res == true && memIsUN != None && memIsUN.get == true) memIsCR = Some(true) 
      memIsSN = Some(res)
    }
    memIsSN.get
  }

  var memIsNF: Option[Boolean] = None

  override def isNF = {if (memIsNF == None) memIsNF = Some(super.isNF); memIsNF.get}

  var memIsCR: Option[Boolean] = None

  override def isCR = {if (memIsCR == None) memIsCR = Some(super.isCR); memIsCR.get}

  var memGetWcrCounterexamples: List[(Node[T], Node[T])] = null

  override def getWcrCounterexamples = {
    if (memGetWcrCounterexamples == null) memGetWcrCounterexamples = super.getWcrCounterexamples
    memGetWcrCounterexamples
  }

  var memGetTransitivePath: List[(Node[T], Node[T])] = null

  override def getTransitivePath = {
    if (memGetTransitivePath == null) memGetTransitivePath = super.getTransitivePath
    memGetTransitivePath
  }

  var memGetNF: List[Node[T]] = null

  override def getNF = {
    if (memGetNF == null) memGetNF = super.getNF
    memGetNF
  }

  var memGetJoiningNodes = getFunctionHashMap(getJoiningNodes)

  override def getJoiningNodes(node: Node[T]) = memGetJoiningNodes.getOrElseUpdate(node, super.getJoiningNodes(node))

  var memGetCrCounterexamples: List[(Node[T], Node[T])] = null

  override def getCrCounterexample = {
    if (memGetCrCounterexamples == null) memGetCrCounterexamples = super.getCrCounterexample
    memGetCrCounterexamples
  }

  var memCycles: List[List[(Node[T], Node[T])]] = null

  override def cycles = {
    if (memCycles == null) memCycles = super.cycles
    memCycles
  }

  var memClosure: List[Node[T]] = null

  override def closure = {
    if (memClosure == null) memClosure = super.closure
    memClosure
  }

  @deprecated("Use addEdge in DirectedGraph instead")
  override def -=(node: Node[T]) = super.-=(node)

  @deprecated("Use removeEdge in DirectedGraph instead")
  override def +=(node: Node[T]) = super.+=(node)
}

object RichNode {
  def paintF(n: Node[GUIPayload]) = {
    //fill(if (n == p.currentNode) 60 else 0)
    p.fill(0)
    p.ellipse(n.pos.x, n.pos.y, 30f, 30f)
    p.fill(255)
    if (n.isNF) p.fill(GREEN)
    p.text(n.name, n.pos.x, n.pos.y)
    p.fill(255)
    for (i <- n.connections)
      arrow(n.pos.x, n.pos.y, i.pos.x, i.pos.y, 15)
  }
}

/**
 * Implements Methods which are no integral Part of Node
 * For Example drawing a Node on Display etc.
 * Usually used via implicit typeconversion
 */
class RichNode(val n: Node[GUIPayload]) {
  def paint(): Unit = paint(List[(Node[GUIPayload], Node[GUIPayload])]())

  def paint(markedPath: List[(Node[GUIPayload], Node[GUIPayload])]) {
    p.fill(0)
    p.ellipse(n.pos.x, n.pos.y, 30f, 30f)
    p.fill(255)
    if (n.isNF) p.fill(GREEN)
    p.text(n.name, n.pos.x, n.pos.y - 5)
    p.fill(255)
    for (i <- n.connections) {
      if (markedPath contains (n, i)) p.stroke(RED)
      arrow(n.pos.x, n.pos.y, i.pos.x, i.pos.y, 15)
      p.stroke(255)
    }
  }
}

object RichNodePreamble {
  implicit def nodeToRichNodeWrapper(n: Node[GUIPayload]) = new RichNode(n)
}