import java.awt.event.KeyEvent
import processing.core._
import processing.core.PApplet._
import processing.core.PConstants._
import PointPreamble._
import GUIPayloadPreamble._
import RichNodePreamble._
import ProcessingWrapper._

object DragMode extends Enumeration {
  val NoDrag, Delete, CreateEdge = Value
}

import DragMode._

object ARG {
  var args: Array[String] = null

  def setArgs(ar: Array[String]) {args = ar}
}

object AWTToSwingWrapper {
  def apply(from: java.awt.Component): swing.Component = {
    class InnerWrapper extends swing.Component {
      peer.add(from)
    }
    new InnerWrapper()
  }
}

class AbstractRewriteTool extends PApplet {
  ProcessingWrapper.setParent(this)

  val RANDOMNODES = 10

  var graph = DirectedGraph.getRandomGraph(RANDOMNODES, 1)
  var guiGraph = new GUIDirectedGraph(graph, XSIZE - XBORDER, YSIZE - XBORDER)

  val messageSystem = MessageSystem(this)
  var gameDialog: GameDialog = null
  var textInput = new TextInput()(parsePrompt)
  val forceTimed = new LimitedTimeCall(3000)

  var currentNode: Node[GUIPayload] = null
  var mousePressedPosition: Point = (0, 0)
  var dragLine = Line.zeroLine // Line which results from pressing ans dragging the mouse

  var dragMode = NoDrag
  var showFPS = false
  var gameMode = false
  var showHelp = false

  def resetGuiGraph() {
    guiGraph = new GUIDirectedGraph(graph, XSIZE - XBORDER, YSIZE - XBORDER)
    textInput = new TextInput()(parsePrompt)
  }

  def resetGraphSystem() {
    graph = new MemDirectedGraph()
    resetGuiGraph()
  }

  def parsePrompt(s: String) {
    if (s != "") {
      ARTStringParser(s) match {
        case None =>
          if (s != "") {
            if (s.length < 3) {
              graph += (s, Nil)
              val rnd = scala.util.Random
              graph(s).pos = (mouseX + (rnd.nextInt(10) - 5) max 0, mouseY + (rnd.nextInt(10) - 5) max 0)
              forceTimed.runFor(50)
            }
            else
              messageSystem.addMessage("Error while parsing Prompt: The name of a node was longer than 2 symbols", 10, RED)

          }
        case Some(p) =>
          if (ARTParser.resultLengthOk(p)) {
            val x = p(0)
            graph += x
            forceTimed.runFor(300)
          } else {
            messageSystem.addMessage("Error while parsing Prompt: The name of a node was longer than 2 symbols", 10, RED)
          }
      }
    }
  }

  def gotoGameMode() {
    if (graph.size == 0) {
      messageSystem.addMessage("Can't enter Game Mode. No Graph!", 10, RED)
    } else {
      guiGraph.markedPath = Nil
      guiGraph.markedNode = Map.empty
      guiGraph.selectedProperty = (null, "")
      guiGraph.colorBackgroundProperty = Map.empty
      guiGraph.hideAllProperties()
      //guiGraph.markHiddenProperty()
      gameDialog = new GameDialog(this)
      gameDialog.refresh()
      //gameDialog.active = false
      messageSystem.clear()
      messageSystem.addMessage("Entering Game Mode", 2)
      gameMode = true
      NODE_NF_TEXT_COLOR = p.color(0) // Set normal forms to black
    }
  }

  override def setup = {
    size(XSIZE, YSIZE, JAVA2D)
    background(0)
    textMode(MODEL)
    textAlign(CENTER, CENTER)
    textSize(standardTextSize)
    strokeWeight(1)

    if(frame!=null) frame.setTitle("Abstract Rewrite Tool")

    smooth()
    //hint(ENABLE_NATIVE_FONTS)
    NODE_NF_TEXT_COLOR = p.color(255, 187, 0)
  }

  var helpText = "Press H for help"
  def switchHelpStatus() = {
    if(showHelp) {
      showHelp = false
      helpText = "Press H for help"
    } else {
      showHelp = true
      helpText = "Press H to exit help"
    }
  }

  override def draw = {
    def showHelpInfo() = {
      p.textSize(10)
      p.fill(BLACK)
      p.textAlign(LEFT, TOP)
      p.text(helpText,0,5)
      p.textAlign(CENTER, CENTER)
      p.textSize(standardTextSize)
    }

    def drawHelpImage() = {
      if(showHelp) {
        p.fill(WHITE)
        p.stroke(WHITE)
        p.rect(0,0,XSIZE,YSIZE)
        if(gameMode)
          p.image(gameHelpImage,XSIZE/2,YSIZE/2)
        else
          p.image(editHelpImage,XSIZE/2,YSIZE/2)
      }
    }

    if (gameMode) {
      background(GAMEMODECOLOR)
      gameDialog.draw()
      guiGraph.draw()
    } else {
      background(EDITMODECOLOR)
      guiGraph.draw()
      //stroke(255)
      if (dragMode == Delete) {
        stroke(LINECOLOR)
        ProcessingWrapper.line(dragLine, LINECOLOR)
        image(scissorsImg, dragLine.p1.x, dragLine.p1.y)
      }
      else if (dragMode == CreateEdge) {
        stroke(LINECOLOR)
        ProcessingWrapper.arrow(dragLine, 0)
      }
    }
    // Draw things that are not dependant on the mode the program is in
    if (showFPS) text(frameRate, 100, 50)
    forceTimed {
      for (i <- 1 to 6) guiGraph.forceDirectedPlacement.iterate()
    }
    if (currentNode != null) {
      currentNode.pos = (mouseX, mouseY)
      forceTimed.runFor(500)
    }

    guiGraph.drawNodeProperties((XSIZE - 200, 20))
    messageSystem.draw()
    textInput.draw()
    drawHelpImage()
    showHelpInfo()
  }


  override def mouseClicked: Unit = {
    if (gameMode) {
      val p = Point(mouseX, mouseY)
      val n = guiGraph.getNodeOnPropertyPosition(p)
      val prop = guiGraph.getSelectedProperty(p)
      if (prop != "none" && prop != "Name" && (guiGraph.hiddenProperties contains (n, prop))) {
        gameDialog.startStopWatch() // Actually only effects the first time called
        if (mouseButton == LEFT) { // Means the User thinks this property is true
          if (gameDialog.correctAnswer(n, prop))
            gameDialog.rightAnswer(n, prop)
          else {
            gameDialog.yesWrongAnswer(n, prop)
          }
        } else if (mouseButton == RIGHT) { // Means the User thinks this property is false
          if (!gameDialog.correctAnswer(n, prop))
            gameDialog.rightAnswer(n, prop)
          else {
            gameDialog.noWrongAnswer(n, prop)
          }
        }
        guiGraph.showNodeProperty(n, prop)
        if (guiGraph.hiddenProperties.isEmpty) gameDialog.stopStopWatch()
      }
    }
  }

  override def mouseReleased() = {
    if (gameMode) {
    } else {
      if (dragMode == Delete) {
        // Remove nodes if intersected
        for (n <- graph) if (dragLine intersectsCircle (n.pos, guiGraph.nodeRadius)) graph -= n
        // Remove selfedges if intersected
        for (n <- graph filter {x => x.connections contains x}) {
          if (dragLine intersectsSelfEdge (n.pos, guiGraph.nodeRadius))
            graph.removeEdge(n, n)
        }
        val lines = (for (n1 <- graph; n2 <- n1) yield ((n1, n2), Line(n1, n2))).toList.distinct
        val deleteLines = lines filter (n => n._2 intersects dragLine)
        deleteLines foreach {l => graph.removeEdge(l._1._1, l._1._2)}
        forceTimed.runFor(500)
      } else if (dragMode == CreateEdge) {
        val startingNode = guiGraph.getNodeAtPosition(dragLine.p1.x, dragLine.p1.y, guiGraph.nodeRadius)
        val endingNode = guiGraph.getNodeAtPosition(dragLine.p2.x, dragLine.p2.y, guiGraph.nodeRadius)
        if (endingNode != null) graph.addEdge(startingNode, endingNode)
        forceTimed.runFor(500)
      }
      dragLine = Line.zeroLine
      dragMode = NoDrag
    }
    currentNode = null
  }

  override def mousePressed() = {
    if (gameMode) {
      if (mouseButton == LEFT) {
        currentNode = guiGraph.getNodeAtPosition(mouseX, mouseY, 17)
      }
    } else {
      if (mouseButton == LEFT) {
        if (textInput.visible) textInput += RETURN // It is possible to accept an Input by clicking
        currentNode = guiGraph.getNodeAtPosition(mouseX, mouseY, 17)
        guiGraph.drawExampleForSelectedProperty(mouseX, mouseY)
      } else if (mouseButton == RIGHT) {
        mousePressedPosition = (mouseX, mouseY)
        dragMode = if (guiGraph.getNodeAtPosition(mouseX, mouseY, guiGraph.nodeRadius) == null) Delete else CreateEdge
      }
    }
  }

  override def mouseDragged() = {
    if (gameMode) {

    } else {
      if (mouseButton == RIGHT) {
        dragLine = Line(mousePressedPosition, (mouseX, mouseY))
      } else if (currentNode != null) {
        currentNode.pos = (mouseX, mouseY)
      }
    }
  }

  override def keyTyped = {
    def graphNodelengthOk(g: DirectedGraph): Boolean = { // Checks if all names of nodes of a graph are 2 symbols maximum
      def ok(s: String) = s.length < 3
      g.forall(x => ok(x.name) && x.connections.forall(y => ok(y.name)))
    }
    def openDialogFunction(file: String) {
      DirectedGraph.getGraphFromFile(file) match {
        case Some(g) =>
          if (graphNodelengthOk(g)) {
            graph = g
            resetGuiGraph()
            messageSystem.addMessage("Loading " + file + " successful", 3, GREEN)
            forceTimed.runFor(1000)
          } else {
            messageSystem.addMessage("Error while parsing File: The name of a node was longer than 2 symbols", 10, RED)
          }
        case _ => messageSystem.addMessage("Error while parsing File: " + file, 10, RED)
      }
      if (gameMode) gotoGameMode()
      this.requestFocus()
    }
    def sParseDialogFunction(s: String) {
      DirectedGraph.getGraphFromString(s) match {
        case Some(g) =>
          if (graphNodelengthOk(g)) {
            graph = g
            resetGuiGraph()
          } else {
            messageSystem.addMessage("Error while parsing Text: The name of a node was longer than 2 symbols", 10, RED)
          }
        case None => messageSystem.addMessage("Error while parsing Text", 10, RED)
      }
      if (gameMode) gotoGameMode()
      this.requestFocus()
    }

    if (textInput.visible) {
      textInput += key
    }
    else {
      if (gameMode) {
        key match {
          case 'h' => switchHelpStatus()
          case 'g' =>
            guiGraph.showAllProperties()
            guiGraph.colorBackgroundProperty = Map.empty
            messageSystem.clear()
            messageSystem.addMessage("Entering Editing Mode", 2)
            gameMode = false
            NODE_NF_TEXT_COLOR = p.color(255, 187, 0) // Set normal forms to orange
          case 'r' =>
            guiGraph.forceDirectedPlacement.t = 300
            forceTimed.runFor(1000)
          case 'n' =>
            graph = DirectedGraph.getRandomGraph(RANDOMNODES, 1)
            resetGuiGraph()
            gotoGameMode()
          case 'l' => SwingFileDialog("./", "Open")(openDialogFunction)
          case 'p' => SParseDialog(graph.toString)(sParseDialogFunction)
          case _ => messageSystem.addMessage("Unknown key pressed: " + key, 5, RED)
        }
      }
      else {
        key match {
          case 'h' => switchHelpStatus()
          case 'g' => gotoGameMode()
          case 'r' => {
            guiGraph.forceDirectedPlacement.t = 300
            forceTimed.runFor(1000)
          }
          case 'c' => resetGraphSystem()
          case 'n' =>
            graph = DirectedGraph.getRandomGraph(RANDOMNODES, 1)
            resetGuiGraph()
          case ENTER => textInput.visible = true
          case 'l' => SwingFileDialog("./", "Open")(openDialogFunction)
          case 's' =>
            SwingFileDialog("./", "Save") {
              import java.io.FileWriter
              file =>
                val w = new FileWriter(file)
                w.write(graph.toString)
                w.close()
            }
          case 'f' => showFPS = !showFPS
          case 'o' => println(guiGraph)
          case 'p' => SParseDialog(graph.toString)(sParseDialogFunction)
          case 'k' => println(guiGraph.gotIntersections)
          case 'x' => SParseDialog(LatexOutput.getLatexSource(guiGraph),"LaTeX Output"){x=>()}
          case _ => messageSystem.addMessage("Unknown key pressed: " + key, 5, RED)
        }
      }
    }
  }
}

object Main {
  def main(args: Array[String]) {
    ARG.setArgs(args)
    val a = PApplet.main(List("AbstractRewriteTool").toArray[String])
  }
}